/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
/*
 * Copyright 2004 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <netinet/in.h>
#include <inttypes.h>
#include <zlib.h>

#define COMPRESSED 1
#define UNCOMPRESSED 0

#define DEV_BSIZE 512
#define BLKHDR 1
#define SIGNATURE "compzlib"

int
main(int argc, char *argv[]) 
{
	char *tfname1, *tfname2, *cmdline;
	uint32_t blksize, count, dev_bsize;
	uint32_t *index, offbase;
	uLongf cblksize;
	int fd, tfd, rv;
	unsigned char *data_block, *compressed_block;
	ssize_t rbytes, wbytes, lastread;
	uint32_t wint;
	char sign[10];

	if (argc < 3) {
		printf("Usage: decompress <filename> <filename>\n\n");
		exit(1);
	}

	tfname1 = argv[1];
	tfname2 = argv[2];
	dev_bsize = DEV_BSIZE;

	fd = open(tfname1, O_RDONLY, 0);
	if (fd == -1) {
		perror("decompress[1]: ");
		exit(1);
	}

	if (read(fd, &sign, 8) == -1) {
		perror("decompress[2]: ");
		close(fd);
		exit(1);
	}

	sign[8] = '\0';
	if (strcmp(SIGNATURE, sign) != 0) {
		fprintf(stderr, "Header signature incorrect\n");
		close(fd);
		exit(1);
	}

	/*
	if (read(fd, &dev_bsize, 4) == -1) {
		perror("decompress[2]: ");
		close(fd);
		exit(1);
	}
	dev_bsize = ntohl(dev_bsize);
	*/

	if (read(fd, &blksize, 4) == -1) {
		perror("decompress[3]: ");
		close(fd);
		exit(1);
	}
	blksize = ntohl(blksize);

	if (read(fd, &count, 4) == -1) {
		perror("decompress[3]: ");
		close(fd);
		exit(1);
	}
	count = ntohl(count);

	if (read(fd, &lastread, 4) == -1) {
		perror("decompress[3a]: ");
		close(fd);
		exit(1);
	}
	lastread = ntohl(lastread);

	cblksize = blksize + (blksize >> 6);
	compressed_block = (unsigned char *)malloc(blksize + BLKHDR);
	data_block = (unsigned char *)malloc(cblksize);

	if (compressed_block == NULL || data_block == NULL) {
		perror("decompress[4]: ");
		exit(1);
	}

	tfd = open(tfname2, O_WRONLY | O_CREAT | O_SYNC | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (tfd == -1) {
		perror("decompress[5]: ");
		exit(1);
	}

	index = malloc(sizeof(uint32_t) * count);
	if (index == NULL) {
		perror("decompress[6]: ");
		close(fd);
		exit(1);
	}

	printf("Index will have %u entries\n", count);

	for (wint=0; wint < count; wint++) {
		if (read(fd, index+wint, 4) == -1) {
			perror("decompress[7]: ");
			close(fd);
			close(tfd);
			exit(1);
		}
		index[wint] = ntohl(index[wint]);
	}
	
	offbase = tell(fd);

	for (wint=0; wint < (count-1); wint++) {
		unsigned long blksz;

		if (lseek(fd, index[wint] + offbase, SEEK_SET) == -1) {
			perror("decompress[8]: ");
			close(fd);
			close(tfd);
			exit(1);
		}
		wbytes = index[wint+1] - index[wint];
		rbytes = read(fd, compressed_block, wbytes);
		if (rbytes == -1) {
			perror("decompress[9]: ");
			close(fd);
			close(tfd);
			exit(1);
		} else if (rbytes < wbytes) {
			fprintf(stderr, "Block %u, requested %u bytes, read %u bytes\n", wint, wbytes, rbytes);
			close(fd);
			close(tfd);
			exit(1);
		}

		/*if (count - 1 - wint == 1) {
			blksz = lastread;
		} else {
			blksz = blksize;
		}*/
		blksz = cblksize;

		if (*compressed_block == UNCOMPRESSED) {
			memcpy(data_block, compressed_block+BLKHDR, rbytes-BLKHDR);
			blksz = rbytes-BLKHDR;
printf("Blk %u, UNCOMPRESSED, blksz = %u\n", wint, blksz);

		} else {
			/*lzjb_decompress(compressed_block+BLKHDR, data_block, rbytes-BLKHDR, blksz);*/
			rv = uncompress(data_block, &blksz, compressed_block+BLKHDR, rbytes-BLKHDR);
			if (rv == Z_BUF_ERROR) {
				fprintf(stderr, "Zlib error: Decompress buffer too small\n");
			} else if (rv == Z_DATA_ERROR) {
				fprintf(stderr, "Zlib error: Compressed data corrupted wint = %u, len = %u, pos = %u\n", wint, rbytes-BLKHDR, index[wint]);
			} else if (rv == Z_MEM_ERROR) {
				fprintf(stderr, "Zlib error: Out of memory\n");
			}
			if (rv != Z_OK) {
				close(fd);
				close(tfd);
				exit(1);
			}
		}
		wbytes = write(tfd, data_block, blksz);
		if (wbytes == -1) {
			perror("decompress[10]: ");
			close(fd);
			close(tfd);
			exit(1);
		} else if (wbytes != blksz) {
			fprintf(stderr, "Block %u, to write %u bytes, written %u bytes\n", wint, rbytes, wbytes);
			close(fd);
			close(tfd);
			exit(1);
		}
	}
	close(fd);
	close(tfd);

	return (0);
}
